unit Root;
{$I Common.inc}
//************************************************************************
//                                                         *
//                                                                       *
//   :                                             *
// TItem --                                                       *
// TSet  --                                          *
// TCostSet --                    *
//************************************************************************

interface

type

  //    

  TCompare = (cmpEq,     //   
              cmpLess,   // 
              cmpGreate, // 
              cmpIncomp  //  (Incomparable)
             );

  //   "" --   

  TItem = class (TObject)
    public
      function Compare(arg: TItem): TCompare; virtual;
      function Copy: TItem;  virtual;{ abstract;}
      procedure Print(var aFile: TextFile); virtual; abstract;
      procedure Expo;
  end;

  //  " "

  TSet = class (TItem)
    protected
      mCount: Longint;  //   
      procedure Clr(aDestroy: boolean); virtual; abstract;
    public
      //    :
      procedure Add(arg : TSet); virtual; abstract;
      procedure Sub(arg : TSet); virtual; abstract;
      procedure Mul(arg : TSet); virtual; abstract;
      procedure ExOr(arg : TSet); virtual; abstract;
      function  TestIntersect(arg: TSet): boolean;
      function  Compare(arg: TItem): TCompare; override;
      function  Exist(arg : TItem): boolean; virtual; abstract;
      // 
      procedure Clear;          //    
      procedure ClrAndDestroy;  //    
      // , ,  
      function  Insert(arg: TItem): boolean; virtual; abstract;
      procedure Delete(arg: TItem); virtual; abstract;
      procedure CopyItems(arg: TSet);
      //   
      function  GetFirst: TItem; virtual; abstract;
      function  GetNext: TItem; virtual; abstract;
      //      
      function  GetItem(index: integer): TItem;
      function  GetObject(aItem: TItem): TItem;
      // -  
      procedure PositionPush;  virtual; abstract;
      procedure PositionPop;  virtual; abstract;
      //  
      procedure CoverToDissect; //    
      function  GetCount: integer;
      function  Copy: TItem; override;
      procedure Print(var aFile: TextFile); override;
      destructor Destroy;  override;
  end;

  //  "  ",   
  //   ()   ()

  TCostSet = class (TItem)
    protected
      procedure Clr(aDestroy: boolean);
    public
      mCost: integer;      // ,  
      mSet : TSet;         //   
      mFlag  : boolean;    //   
      mDestroy : boolean;  //    
      constructor Create(aCost: integer; aSet: TSet; aDestroy : boolean);
      constructor CreateEmpty;
      destructor  Destroy;  override;
      function    Copy: TItem; override;
      procedure   Clear;
      procedure   ClrAndDestroy;
      function    Compare(arg: TItem): TCompare; override;
      procedure   Print(var aFile: TextFile); override;
      procedure   Append(arg : TCostSet);
      procedure   Insert(arg : TCostSet);
  end;

//    

function CreateSet: TSet;

//////////////////////////////////////////////////////////////////////////
implementation
uses Items, SetList, Graph;
//////////////////////////////////////////////////////////////////////////

//    

function CreateSet: TSet;
begin
  Result:= TSetList.Create;
end;

//////////////////////////////////////////////////////////////////////////
// TItem   --   "" --   
//////////////////////////////////////////////////////////////////////////

function TItem.Compare(arg: TItem): TCompare;
begin
  if arg = Self
    then Result:= cmpEq
    else Result:= cmpIncomp;
end;


function TItem.Copy: TItem;
begin
  Result:= nil
end;

procedure TItem.Expo;       //    
begin
  if Assigned(Self) then Print(Output);
end;
//////////////////////////////////////////////////////////////////////////
// TSet  --   ""
//////////////////////////////////////////////////////////////////////////

//   

function TSet.Compare(arg: TItem): TCompare;
var p, q : TItem;
begin
  Result:= cmpEq;
  if Self=arg then Exit;  //     

  //    ,   

  if not (arg is TSet) then begin
    Result:=cmpIncomp;    // 
    Exit;
  end;

  //   

  PositionPush;             // .    
  (arg as TSet).PositionPush;  // .   
  p:= GetFirst;             //      
  q:= (arg as TSet).GetFirst;  //     
  while Assigned(p) and Assigned(q) do begin  //   
    Result:= p.Compare(q);                    //  
    if Result <> cmpEq then Break;            // ,   
    p:= GetNext;  q:=(arg as TSet).GetNext;   //  
  end;
  if Result=cmpEq then begin               //   . 
    if GetCount < (arg as TSet).GetCount
      then Result:= cmpLess                //   
    else if GetCount > TSet(arg).GetCount
      then Result:= cmpGreate              //   
  end;
  (arg as TSet).PositionPop;    // .   
  PositionPop;              // .    
end;

//   

function TSet.Copy: TItem;
var p, q : TItem;
begin
  Result:= CreateSet;
  PositionPush;
  p:= GetFirst;
  while Assigned(p) do begin
    if p is TSet
      then q:= p.Copy          //  
      else q:= p;              //  
    (Result as TSet).Insert(q);
    p:= GetNext;
  end;
  PositionPop;
end;

//       

procedure TSet.CopyItems(arg: TSet);
var t: TItem;
begin
  Clear;     //     
  arg.PositionPush;
  t:= arg.GetFirst;
  while Assigned(t) do begin
    Self.Insert(t);
    t:= arg.GetNext;
  end;
  arg.PositionPop;
end;

//    

procedure TSet.Clear;
begin
  Clr(false)
end;

//    

procedure TSet.ClrAndDestroy;
begin
  Clr(true)
end;

// 

destructor TSet.Destroy;
begin
  Clear;   //    
  inherited;
end;

//    .
//     TCostSet!

procedure TSet.CoverToDissect;
var   i, j : integer;  // 
    Si, Sj : TCostSet; //  
begin
  for i:= 1 to GetCount-1 do begin
    Si:= GetItem(i) as TCostSet;
    //      
    for j:= i+1 to GetCount do begin
      Sj:= GetItem(j) as TCostSet;
      Si.mSet.Sub(Sj.mSet);
    end;
  end;
end;

//     .
//   PrintLevel  OldFlag
//    
//    

var PrintLevel : integer = 0;
    OldFlag: boolean;

procedure TSet.Print(var aFile: TextFile);
var t : TItem;      //  
    S : string;     //   
    i : integer;
    flag: boolean;
begin
  S:=''; for i:=1 to 2*PrintLevel do S:= S+#32;
  Write(aFile,S+'{');
  if PrintLevel=0 then OldFlag:= true;
  Inc(PrintLevel);
  flag:= false;
  PositionPush;
  t:= GetFirst;
  while Assigned(t) do begin
    flag:= (t is TSet) or
           (t is TCostSet) or
           (t is TGraph) or
           (t is TNode);
    if flag and OldFlag then Writeln(aFile);
    t.Print(aFile);
    t:= GetNext;
  end;
  PositionPop;
  if flag
    then Writeln(aFile,S+'} : ', mCount)
    else Writeln(aFile,' } : ', mCount);
  OldFlag:= flag;
  Dec(PrintLevel);
end;

//    

function TSet.TestIntersect(arg: TSet): boolean;
var  Temp: TSet;   //   
begin
  Temp:= CreateSet;
  Temp.CopyItems(Self);
  Temp.Mul(arg);
  Result:= Temp.GetCount > 0;
  Temp.Free;
end;

function TSet.GetCount: integer;
begin
  Result:= mCount;
end;

//       
// index = 1..GetCount

function TSet.GetItem(index: integer): TItem;
begin
  Result:= nil;
  if (index<1) or (index>GetCount) then exit;
  PositionPush;             // . 
  Result:= GetFirst;        //  
  while index>1 do begin    //   
    Dec(Index);
    Result:= GetNext;
  end;
  PositionPop;              // . 
end;

//      

function TSet.GetObject(aItem: TItem): TItem;
begin
  PositionPush;            // . 
  Result:= GetFirst;       //  
  while Assigned(Result) and
        (Result.Compare(aItem) <> cmpEq)
    do Result:= GetNext;   //   
  PositionPop;             // . 
end;

//////////////////////////////////////////////////////////////////////////
// TCostSet --   
//////////////////////////////////////////////////////////////////////////

constructor TCostSet.Create(aCost: integer; aSet: TSet; aDestroy : boolean);
begin
  inherited Create;
  mCost:= aCost;     // 
  mSet := aSet;      //   
  if not Assigned(mSet) then begin
    mDestroy:= true;
    mSet:= CreateSet;
  end;
  if aDestroy then mDestroy:= true;  //    
end;

constructor TCostSet.CreateEmpty;
begin
  inherited Create;
  mDestroy:= true;
  mSet:= CreateSet;
end;

function TCostSet.Copy: TItem;
begin
  Result:= TCostSet.Create(mCost, mSet.Copy as TSet, mDestroy);
  (Result as TCostSet).mDestroy:= true;
end;

destructor TCostSet.Destroy;
begin
  if mDestroy then mSet.Free;
  inherited;
end;


procedure TCostSet.Clr(aDestroy: boolean);
begin
  if aDestroy
    then mSet.ClrAndDestroy
    else mSet.Clear;
  mCost:=0;
end;

procedure TCostSet.Clear;
begin
  Clr(false)
end;

procedure TCostSet.ClrAndDestroy;
begin
  Clr(true)
end;

//  

procedure TCostSet.Append(arg: TCostSet);
var CS : TCostSet;
     i : integer;
begin
  if not Assigned(arg) then Exit;
  for i:= 1 to arg.mSet.GetCount do begin
    CS:= TCostSet(arg.mSet.GetItem(i));
    Insert(CS);
  end;
end;

//    ()   

procedure TCostSet.Insert(arg: TCostSet);
begin
  if not Assigned(arg) then Exit;
  Inc(mCost, arg.mCost);
  mSet.Insert(arg);
end;

//    :

function TCostSet.Compare(arg: TItem): TCompare;
begin
  if Self = arg then begin
    Result:= cmpEq;
    Exit;
  end;
  Result:= cmpIncomp;  //  
  if not Assigned(arg) then Exit;
  if not (arg is TCostSet) then Exit;
  //  :
  if mCost = (arg as TCostSet).mCost then begin
    //   
    //   
    if mSet.GetCount > (arg as TCostSet).mSet.GetCount
      then Result:= cmpLess
    else if mSet.GetCount < (arg as TCostSet).mSet.GetCount
      then Result:= cmpGreate
      //      
      else Result:= mSet.Compare((arg as TCostSet).mSet);
  end else begin
    //    ,
    //     ( -  )
    if mCost * (arg as TCostSet).mSet.GetCount <
      (arg as TCostSet).mCost * mSet.GetCount
      then  Result:= cmpLess
      else
        if mCost * (arg as TCostSet).mSet.GetCount >
            (arg as TCostSet).mCost * mSet.GetCount
          then   Result:= cmpGreate
          else  {   - }
            if mSet.GetCount > (arg as TCostSet).mSet.GetCount
              then Result:= cmpLess
              else
                if mSet.GetCount < (arg as TCostSet).mSet.GetCount
                  then Result:= cmpGreate
  end;
end;


procedure TCostSet.Print(var aFile: TextFile);
var t: TItem;
begin
  t:= mSet.GetItem(1);
  if Assigned(t) and (t is ClassType) then begin
    Writeln(aFile, 'Cost= ', mCost);  //     
  end else begin
    Write(aFile, ' Cost=', mCost:4, '  ');  //   
  end;
  mSet.Print(aFile)
end;

//////////////////////////////////////////////////////////////////////////

end.
